# -*- coding: utf-8 -*-
from djeese.apps import AppConfiguration
from djeese.commands import BaseCommand, CommandError
from djeese.input_helpers import ask, ask_password, ask_boolean
from djeese.printer import Printer
import os
import requests
import shutil
import subprocess
import tempfile

TARGET_URL = 'https://control.djeese.com/api/v1/apps/upload/'
LOGIN_URL = 'https://control.djeese.com/api/v1/login/'

AUTH_FILE = os.path.join(os.path.expanduser('~'), '.djeese')

class Command(BaseCommand):
    help = 'Upload an app.'

    def handle(self, setupfile=None, appfile=None, **options):
        if not setupfile:
            raise CommandError("You must provide the path to your apps setup.py as first argument")
        if not appfile:
            raise CommandError("You must provide the path to your app file as first argument")
        if not os.path.exists(setupfile):
            raise CommandError("Could not find setup.py at %r" % setupfile)
        if not os.path.exists(appfile):
            raise CommandError("Could not find setup.py at %r" % appfile)
        username, password = self.get_auth()
        distdir = tempfile.mkdtemp(prefix='djeese')
        try:
            self.run(setupfile, appfile, distdir, username, password, **options)
        finally:
            shutil.rmtree(distdir)
    
    def get_auth(self):
        username, password = None, None
        if os.path.exists(AUTH_FILE):
            fobj = open(AUTH_FILE)
            try:
                data = fobj.read()
            finally:
                fobj.close()
            if data.count(':') == 1:
                username, password = [bit.strip() for bit in data.split(':')]
        if not (username and password):
            username = ask("Username")
            password = ask_password("Password:")
            if ask_boolean("Save login data?", default=True) == 'true':
                fobj = open(AUTH_FILE, 'w')
                try:
                    data = fobj.write(u'%s:%s' % (username, password))
                finally:
                    fobj.close()
        return username, password
                
    
    def run(self, setupfile, appfile, distdir, username, password, **options):
        fnull = open(os.devnull, 'w')
        try:
            subprocess.check_call(['python', setupfile, 'sdist', '-d', distdir], stdout=fnull)
        finally:
            fnull.close()
        eggfile = os.path.join(distdir, os.listdir(distdir)[0])
        config = AppConfiguration(int(options['verbosity']))
        config.read(appfile)
        config.validate()
        appname = config['app']['name']
        response = self.upload(appfile, appname, eggfile, username, password)
        printer = Printer(int(options['verbosity']))
        if response.status_code == 201:
            printer.always("Upload successful (created)")
        elif response.status_code == 204:
            printer.always("Upload successful (updated)")
        elif response.status_code == 400:
            printer.error(response.content)
            printer.always("Upload failed")
        elif response.status_code == 403:
            printer.error("Authentication failed")
            printer.always("Upload failed")
        elif response.status_code == 502:
            printer.error("Temporarily unavailable")
            printer.always("Upload failed")
        else:
            printer.error("Unexpected response: %s" % response.status_code)
            printer.error(response.content)
            printer.always("Upload failed")
        
    def upload(self, configfile, appname, eggfile, username, password):
        with open(configfile, 'r') as configfileobj:
            configdata = configfileobj.read()
        with open(eggfile, 'rb') as eggfileobj:
            files = {
                'egg': eggfileobj,
            }
            data = {
                'config': configdata,
                'app': appname,
            }
            session = requests.session()
            response = session.post(LOGIN_URL, {'username': username, 'password': password})
            if response.status_code != 204:
                return response 
            response = session.post(TARGET_URL, data=data, files=files)
            return response



from collections import defaultdict
from djeese.apps import AppConfiguration, VALID_TYPES
from djeese.commands import BaseCommand
from djeese.input_helpers import (RegexValidator, ask, SlugValidator, 
    ask_boolean, ask_multi, URLValidator, ask_choice)
from djeese.utils import slugify, get_package_data


def contrib(config, section, key, method, *args, **kwargs):
    """
    Wrapper to ask the user for input using `method` (passing `args` and
    `kwargs` to that method) and if a value is returned, set it in the `config`
    under `section` and `key`.
    """
    value = method(*args, **kwargs)
    if value:
        config[section][key] = value
    return value


class Command(BaseCommand):
    help = 'Create a djeese apps.'

    def handle(self, **options):
        config = AppConfiguration(1)
        letterfirst = RegexValidator(r'^[a-zA-Z]', "Must start with a letter")
        name = contrib(config, 'app', 'name', ask, "Name", letterfirst)
        packagename = contrib(config, 'app', 'packagename', ask, 'Package name on PyPI', SlugValidator(), default=slugify(name))
        check_net = ask_boolean("Should we try to get additional information from djangopackages?", default=True)
        if check_net:
            data = get_package_data(packagename)
        else:
            data = defaultdict(lambda:None)
        contrib(config, 'app', 'private', ask_boolean, "Private")
        contrib(config, 'app', 'url', ask, 'URL', default=data['url'])
        contrib(config, 'app', 'author', ask, 'Author', default=data['author'], required=False)
        contrib(config, 'app', 'author-url', ask, 'Author URL (optional)', default=data['author_url'], required=False)
        contrib(config, 'app', 'installed-apps', ask_multi, "Installed apps", minitems=1)
        contrib(config, 'app', 'version', ask, 'Version', default=data['version'])
        contrib(config, 'app', 'description', ask, 'Description (short)', default=data['description'])
        contrib(config, 'app', 'license', ask, 'License', default=data['license'])
        contrib(config, 'app', 'license-text', ask, 'License Text (URL)', URLValidator(), default=data['license_text_url'])
        contrib(config, 'app', 'translation-url', ask, 'URL to the translation page, eg transifex (optional)', URLValidator(), required=False)
        contrib(config, 'app', 'settings', ask_multi, 'Settings (optional)')
        contrib(config, 'app', 'plugins', ask_multi, 'Plugin (class) names (optional)')
        contrib(config, 'app', 'apphook', ask_multi, 'Apphook (class) names (optional)')
        for setting in config['app'].getlist('settings'):
            contrib(config, setting, 'name', ask, 'Name of the setting %r (Python)' % setting)
            contrib(config, setting, 'verbose-name', ask, 'Verbose name of the setting %r' % setting)
            contrib(config, setting, 'type', ask_choice, 'Type of the setting %r' % setting, choices=VALID_TYPES)
            contrib(config, setting, 'default', ask, 'Default value for setting %r (optional)' % setting, required=False)
            contrib(config, setting, 'required', ask_boolean, 'Is setting %r required' % setting, default=True)
            if config[setting].get('default', None):
                contrib(config, setting, 'editable', ask_boolean, 'Is setting %r editable' % setting, default=True)
        if ask_boolean("Does your application expose templates?", default=True) == 'true':
            while True:
                name = ask('Template path (eg %s/plugin.html)' % config['app']['packagename'])
                url = ask('URL to the source of the template', URLValidator())
                config['templates'][name] = url
                if ask_boolean("Are there more templates?") == 'false':
                    break
        fname = '%s.ini' % packagename
        config.write(fname)



# -*- coding: utf-8 -*-
from collections import defaultdict
import re
import requests
import unicodedata
import urlparse
import xmlrpclib
try:
    import json
except ImportError:
    import simplejson as json
try:
    from requests import async
except RuntimeError:
    class dummy_async(object):
        def get(self, url, **kwargs):
            return requests.get(url, **kwargs)
        
        def map(self, rs):
            return rs
    async = dummy_async()


LICENSE_FILE_CANDIDATES = [
    'license',
    'license.txt',
    'LICENSE',
    'LICENSE.txt',
    'LICENSE.TXT',
]

PYPI_KEYS = [
    'author',
    'author_email',
    'license',
    'version',
]

DJANGOPACKAGES_KEYS = [
    'repo_description',
]

def check_urls(urls, timeout=5):
    """
    Checks a list of URLs (asynchronously if possible)
    """
    rs = [async.get(url, timeout=5) for url in urls]
    return [(response.url, 200 <= response.status_code < 300) for response in async.map(rs)]

def _check_url(url):
    """
    Checks a single URL
    """
    return requests.get(url, timeout=5).status_code == 200

def slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.
    
    Slightly adapted from django.template.default_filters.slugify
    """
    value = unicode(value)
    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore')
    value = unicode(re.sub('[^\w\s-]', '', value).strip().lower())
    return re.sub('[-\s]+', '-', value)

def _transform_pypi(data):
    """
    Transform the data from PyPI into the data we want.
    """
    return dict([(key, value) for key, value in data.items() if value != "UNKNOWN" and key in PYPI_KEYS])

def get_pypi_data(slug):
    """
    Download and process data on PyPI for an app with given slug, if available.
    If not available, returns empty dictionary.
    """
    client = xmlrpclib.ServerProxy('http://pypi.python.org/pypi')
    try:
        releases = client.package_releases(slug)
    except OSError:
        return {}
    if not releases:
        return {}
    release = releases[0]
    data = client.release_data(slug, release)
    return _transform_pypi(data)

def _find_license_url_bitbucket(scheme, netloc, path, tag):
    """
    Try to find the URL to the license text on bitbucket given a bitbucket
    repository URL (urlsplitted) and a version (tag).
    """
    for candidate in LICENSE_FILE_CANDIDATES:
        newpath = '%sraw/%s/%s' % (path, tag, candidate)
        url = urlparse.urlunsplit((scheme, netloc, newpath, '', ''))
        if _check_url(url):
            return url
    return None

def _find_license_url_github(scheme, netloc, path, tag):
    """
    Try to find the URL to the license text on github given a github
    repository URL (urlsplitted) and a version (tag).
    """
    for candidate in LICENSE_FILE_CANDIDATES:
        newpath = '%s/%s/%s' % (path, tag, candidate)
        newnetloc = 'raw.%s' % netloc
        url = urlparse.urlunsplit((scheme, newnetloc, newpath, '', ''))
        if _check_url(url):
            return url
    return None

def _find_license_url(repourl, version):
    """
    Try to find the URL to the license text either on github or bitbucket.
    """
    scheme, netloc, path, _, _ = urlparse.urlsplit(repourl)
    if netloc == 'github.com':
        return _find_license_url_github(scheme, netloc, path, version)
    elif netloc == 'bitbucket.org':
        return _find_license_url_bitbucket(scheme, netloc, path, version)
    else:
        return None
    
def _find_author_url(repourl):
    """
    Try to find the Author URL if the URL is on github or bitbucket.
    """
    scheme, netloc, path, _, _ = urlparse.urlsplit(repourl)
    if netloc in ['github.com', 'bitbucket.org']:
        newpath = '/'.join(path.split('/')[:2])
        return urlparse.urlunsplit((scheme, netloc, newpath, '', ''))
    else:
        return None

def _transform_djangopackages(data):
    """
    Process djangopackages data. 
    """
    url = data.get('repo_url', None)
    version = data.get('pypi_version', None)
    data = dict([(key, value) for key, value in data.items() if key in DJANGOPACKAGES_KEYS])
    data['license_text_url'] = None
    if url and version:
        license_text_url = _find_license_url(url, version)
        if license_text_url:
            data['license_text_url'] = license_text_url
    data['description'] = data.pop('repo_description', None)
    data['url'] = url
    data['author_url'] = _find_author_url(url)
    return data

def get_djangopackages_data(slug):
    """
    Download and process data from djangopackages using their API.
    """
    url = 'http://djangopackages.com/api/v1/package/%s/' % slug
    response = requests.get(url, timeout=5)
    if response.status_code != 200:
        return {'description': None}
    data = json.loads(response.content)
    return _transform_djangopackages(data)

def get_package_data(slug):
    """
    Get package data from Djangopackages and PyPI
    """
    data = defaultdict(lambda:None)
    data.update(get_pypi_data(slug))
    data.update(get_djangopackages_data(slug))
    return data



# -*- coding: utf-8 -*-


class Printer(object):
    """
    Helper object to print certain types of information only if the verbosity
    level given is high enough.
    """
    def __init__(self, verbosity):
        self.verbosity = verbosity
    
    def _print(self, level, message):
        if self.verbosity >= level:
            print message
    
    def info(self, message):
        """
        Print if verbosity is 3 or above
        """
        self._print(3, message)
    
    def warning(self, message):
        """
        Print if verbosity is 2 or above
        """
        self._print(2, message)
    
    def error(self, message):
        """
        Print if verbosity is 1 or above
        """
        self._print(1, message)
        
    def always(self, message):
        self._print(-1, message)
    


# -*- coding: utf-8 -*-
from ConfigParser import SafeConfigParser
from StringIO import StringIO
from djeese.printer import Printer
from djeese.utils import check_urls


VALID_TYPES = ['string', 'stringtuplelist', 'stringlist', 'boolean']
REQUIRED_APP_KEYS = ['name', 'packagename', 'installed-apps', 'description',
                     'license-text', 'license', 'url', 'version', 'private']
EXTRA_APP_KEYS = ['settings', 'author', 'author-url', 'translation-url',
                  'plugins', 'apphooks']
REQUIRED_SETTINGS_KEYS = ['name', 'verbose-name', 'type']
EXTRA_SETTINGS_KEYS = ['default', 'required', 'editable']

def validate_app(config, printer):
    """
    Validate an app (AppConfiguration instance) and print errors using the
    printer (djeese.utils.Printer instance).
    """
    valid = True
    if 'app' not in config:
        printer.error("Section 'app' not found")
        valid = False
    else:
        if not validate_app_section(config, printer):
            valid = False
    if 'templates' in config:
        if not validate_templates_section(config, printer):
            valid = False
    settings = config['app'].getlist('settings')
    for setting in settings:
        if not validate_settings_section(config, printer, setting):
            valid = False
    if valid:
        printer.always("Configuration valid")
    else:
        printer.always("Configuration invalid")
    return valid

def validate_app_section(config, printer):
    """
    Validate the [app] section of an app configuration
    """
    valid = True
    printer.info("Required section 'app' found")
    app = config['app']
    for required in REQUIRED_APP_KEYS:
        if required not in app:
            printer.error("Option '%s' not found in 'app' section" % required)
            valid = False
        else:
            printer.info("Required option '%s' found in 'app' section" % required)
    return valid

def validate_templates_section(config, printer):
    """
    Validate the [templates] section of an app configuration
    """
    valid = True
    templates = config['templates'].as_dict()
    reverse_templates = dict([(v,k) for k,v in templates.items()])
    responses = check_urls(templates.values())
    for url, success in responses:
        name = reverse_templates[url]
        if not success:
            printer.error("Could not load template %r from %r" % (name, url))
            valid = False
        else:
            printer.info("Successfully loaded %r from %r" % (name, url))
    return valid

def validate_settings_section(config, printer, setting):
    """
    Validate the settings section of an app configuration for a setting
    """
    valid = True
    if setting not in config:
        printer.error("Could not find settings section %r" % setting)
        return False
    setting_config = config[setting]
    for required in ['verbose-name', 'name', 'type']:
        if required not in setting_config:
            printer.error("Could not find required option %r in settings section %r" % (required, setting))
            valid = False
    if 'type' in setting_config:
        typevalue = setting_config['type']
        if typevalue not in VALID_TYPES:
            valid_types = ', '.join(VALID_TYPES)
            printer.error("Setting %r type %r is not valid. Valid choices: %s" % (setting, typevalue, valid_types))
    printer.info("Settings section valid")
    return valid

class Section(object):
    """
    Wrapper around ConfigParser to make a nicer interface
    """
    def __init__(self, parser, section):
        self.parser = parser
        self.section = section
    
    def items(self):
        """
        Return the items in this section as list of tuples
        """
        return self.parser.items(self.section)
        
    def as_dict(self):
        """
        Return the items in this section as dictionary
        """
        return dict(self.items())
    
    def get(self, item, default=None):
        """
        Get the item in this section
        """
        if item in self:
            return self[item]
        return default
        
    def getlist(self, item, default=None):
        """
        Get the item in this section as list
        """
        if item in self:
            return [line.strip() for line in self[item].splitlines() if line.strip()]
        if default:
            return default
        return []
    
    def getint(self, item, default=0):
        """
        Get the item in this section as integer
        """
        if item in self:
            return self.parser.getint(self.section, item)
        return default
    
    def getfloat(self, item, default=0.0):
        """
        Get the item in this section as float
        """
        if item in self:
            return self.parser.getfloat(self.section, item)
        return default
    
    def getboolean(self, item, default=False):
        """
        Get the item in this section as boolean
        """
        if item in self:
            return self.parser.getboolean(self.section, item)
        return default
    
    def __contains__(self, item):
        """
        Check if this section contains this item
        """
        return self.parser.has_option(self.section, item)
        
    def __getitem__(self, item):
        """
        Same as self.get, but without the ability to specify a default
        """
        return self.parser.get(self.section, item)
    
    def __setitem__(self, item, value):
        """
        Set an item in this section
        """
        self.parser.set(self.section, item, value)
        
    def __delitem__(self, item):
        """
        Remove an item from this section
        """
        self.parser.remove_option(self.section, item)


class AppConfiguration(object):
    """
    Wrapper around SafeConfigParser to provide a nicer API
    """
    def __init__(self, verbosity=1):
        self.parser = SafeConfigParser()
        self.printer = Printer(verbosity)
        
    def __getitem__(self, item):
        """
        Get a section (Section instance) or create it if it doesn't exist.
        """
        if item not in self:
            self.parser.add_section(item)
        return Section(self.parser, item)
    
    def __contains__(self, item):
        """
        Check if this config has a section
        """
        return self.parser.has_section(item)
    
    def read_string(self, data):
        """
        Read the configuration from a string
        """
        sio = StringIO(data)
        sio.seek(0)
        self.parser.readfp(sio)
    
    def read(self, filepath):
        """
        Read the configuration from a filepath
        """
        self.parser.read(filepath)
    
    def write(self, filepath):
        """
        Write the configuration to a filepath
        """
        if not self.validate():
            return False
        with open(filepath, 'w') as fobj:
            self.parser.write(fobj)
        return True
    
    def validate(self):
        """
        Validate this configuration
        """
        return validate_app(self, self.printer)
